/**
 * Adaptation of the AStar algorithm using Leaflet LatLng objects for distance
 */
import { LatLng } from 'leaflet';

export default function AStar (pathNodes, edges, startNode, endNode) {
  /**
     * Constructor;
     * @param pathNodes A list of nodes on the set path
     * @param edges A set of edges on this map
     * @param startNode The starting location
     * @param endNode The ending location
     */
  this.nodes = pathNodes;
  this.edges = edges;
  this.startNode = startNode;
  this.endNode = endNode;

  this.openSet = [];
  this.closedSet = [];

  function isCross(currentNode, neighbors) {
    return (edge) => {
      const isCrossMap = edge.map.length > 1;
      if (edge.node1._id === currentNode._id) {
        neighbors.push({ node: edge.node2, cross: isCrossMap });
      } else if (edge.node2._id === currentNode._id) {
        neighbors.push({ node: edge.node1, cross: isCrossMap });
      }
    };
  }

  function processNeighbor(currentNode) {
    return (neighborObj) => {
      const neighbor = neighborObj.node;
      // If the neighbor is in the closed set, skip it
      if (this.inSet(this.closedSet, neighbor)) {
        return;
      }

      // Calculate the gScore
      // Get the latlngs
      const currentLatLng = new LatLng(currentNode.lat, currentNode.lng);
      const neighborLatLng = new LatLng(neighbor.lat, neighbor.lng);

      // Cross map nodes have a distance of 0
      let distance;
      if (neighborObj.cross) {
        distance = 0;
      } else {
        distance = currentLatLng.distanceTo(neighborLatLng);
      }

      // Manage gScore calculations
      const calculatedGScore = (currentNode.gScore ? currentNode.gScore : -1.0) + distance;
      let bestScore = false;
      // If the neighbor is not in the open set (we haven't visited it yet), go ahead and add it
      if (!this.inSet(this.openSet, neighbor)) {
        bestScore = true;
        neighbor.hScore = this.calculateHeuristic(neighbor, this.endNode);
        this.openSet.push(neighbor);
      } else if (calculatedGScore < neighbor.gScore) {
        // we found a better gScore
        bestScore = true;
      }

      // If this is the best score we've found, update the neighbor
      if (bestScore) {
        neighbor.cameFrom = currentNode;
        neighbor.gScore = calculatedGScore;
        neighbor.fScore = neighbor.gScore + neighbor.hScore;
      }

      // Check if this is the end to reset its value so we can reconstruct the path,
      // otherwise the endNode field will still have a null cameFrom
      if (this.endNode._id === neighbor._id) {
        this.endNode = neighbor;
      }
    };
  }


  /**
   * Executes the algorithm and finds the most efficient route by updating
   * the values indicating where the Node came from.
   * @return True if valid path exists in the graph, False otherwise
   */
  this.runAlgorithm = function() {
    let currentNode;

    // push startNode into openList
    this.openSet.push(this.startNode);

    while (this.openSet.length > 0) {
      [currentNode] = this.openSet; // start with the first node
      const currentNodeIndex = 0;

      // If this is the final node, simply return true: we are done
      if (currentNode._id === this.endNode._id) {
        return true;
      }

      // Add this to the closed set and remove it from the open set
      this.closedSet.push(currentNode);
      this.openSet.splice(currentNodeIndex, 1);

      // Get a list of neighbors
      const neighbors = [];
      this.edges.forEach(isCross(currentNode, neighbors));

      // Determine which neighbors to explore
      neighbors.forEach(processNeighbor(currentNode));

      this.bubbleSort(this.openSet);
    }

    return false;
  };

  /**
     * Work backwards to generate the set of nodes that lie on the
     * most efficient path. Final path is in order from start to finish
     * @return A list of nodes on the most efficient path from start to finish
     */
  this.reconstructPath = function() {
    // Holds the final path
    const path = [];
    const edgesTraversed = [];

    // Start at the end of the list
    let currentNode = this.endNode;

    let nodeA = currentNode;
    let nodeB = currentNode.cameFrom;

    // While there is a previous node, add this node to the current list and change to the next one
    // Build a list of edges on this path
    while (currentNode.cameFrom !== null) {
      path.unshift(currentNode);
      currentNode = currentNode.cameFrom;

      // For every two nodes, we want the edge between these two nodes
      edgesTraversed.push(this.findEdge(nodeA, nodeB));

      // Reassign them to the next values
      nodeA = nodeB;
      nodeB = currentNode.cameFrom;
    }

    // Add the start node which is skipped because its cameFrom field should be null
    path.unshift(this.startNode);

    return [path, edges];
  };

  /**
     * Finds the first edge between these two nodes
     * @param nodeA Node existing in an edge pair
     * @param nodeB Node existing in an edge pair
     */
  this.findEdge = function(nodeA, nodeB) {
    let foundEdge = null;

    this.edges.forEach((edge) => {
      if ((edge.node1._id === nodeA._id || edge.node2._id === nodeA._id) &&
               (edge.node1._id === nodeB._id || edge.node2._id === nodeB._id)) {
        foundEdge = edge;
      }
    });

    return foundEdge;
  };

  /**
     * Calculates the heuristic based on the nodes latitude and longitude
     * The square of the sum of the difference in latitude squared
     * and difference in longitude squared
     * @param node Node to start from
     * @param neighbor Node to get to
     * @return Number Represents the calculated value for traversing to the neighbor
     * TODO: Account for height?
     */
  this.calculateHeuristic = function(node, neighbor) {
    // Calculate distance using LatLng
    // Get the latlngs
    const currentLatLng = new LatLng(node.lat, node.lng);
    const neighborLatLng = new LatLng(neighbor.lat, neighbor.lng);

    const distance = currentLatLng.distanceTo(neighborLatLng);

    return distance;
  };

  /**
     * Standard bubble sorting algorithm
     * @param set Set of nodes to sort
     * @returns A list of nodes in order by FScore
     */
  this.bubbleSort = function(set) {
    let swapped; // keeps track of if we swapped or not to know when to end

    do {
      swapped = false;
      for (let i = 0; i < set.length - 1; i += 1) {
        if (set[i].fScore > set[i + 1].fScore) {
          // perform swap
          const tmp = set[i];
          set[i] = set[i + 1];
          set[i + 1] = tmp;

          swapped = true;
        }
      }
    } while (swapped);

    return set;
  };

  /**
     * Determines if the given node is in the given set
     * @param set Set of nodes
     * @param node Node to check
     * @return True if node is in set, False otherwise
     */
  this.inSet = function(set, node) {
    let inSet = false;
    // Check if the closed set contains neighbor, and skip it if it does (simply return)
    set.forEach((setNode) => {
      if (node._id === setNode._id) {
        inSet = true;
      }
    });

    return inSet;
  };
}
