import { Mongo } from 'meteor/mongo';
import { Meteor } from 'meteor/meteor';
import { check } from 'meteor/check';
import { toTitleCase } from '../algorithms/strings.js';

export const Maps = new Mongo.Collection('maps');

if (Meteor.isServer) {
  // Publish the set of available maps
  Meteor.publish('maps', function tasksPublication() {
    // Set all of the edges to unselected for every map
    const allMaps = Maps.find({});
    allMaps.forEach((map) => {
      map.edges.forEach((edge) => {
        edge.selected = false;
      });
      Meteor.call('maps.highlightEdges', map._id, map.edges);
    });

    return Maps.find({});
  });
}

/**
 * Defines possible API calls to interact with the database
 */
Meteor.methods({

  /**
     * Adds a new map to the database with the criteria from AddMapPopup entered by the user
     * @param mapName Common name of the map
     * @param mapWidth Width of the map image in pixels
     * @param mapHeight Height of the map image in pixels
     * @param mapScale Scale of the map in feet
     * @param mapImageID ID of the image to match 'mapimages' collection
     */
  'maps.addNewMap'(mapName, mapWidth, mapHeight, mapScale, mapImageID) {
    check(mapName, String);
    check(mapWidth, Number);
    check(mapHeight, Number);
    check(mapScale, Number);
    check(mapImageID, String);

    Maps.insert({
      _id: Random.id(),
      name: mapName,
      image: {
        _id: mapImageID, // Same ID from the 'mapimages' collection from file upload
        width: mapWidth,
        height: mapHeight,
      },
      scale: mapScale,
      updated: Date.parse(Date()),
      nodes: [],
      edges: [],
    });
  },

  /**
     * Updates the map properties with the new information
     * @param mapID Mongo assigned ID of the map
     * @param mapName Common name of the map
     * @param mapWidth Width of the map image in pixels
     * @param mapHeight Height of the map image in pixels
     * @param mapScale Scale of the map in feet
     */
  'maps.updateMap'(mapID, mapName, mapWidth, mapHeight, mapScale) {
    check(mapID, String);
    check(mapName, String);
    check(mapWidth, String);
    check(mapHeight, String);
    check(mapScale, String);

    const height = parseInt(mapHeight);
    const width = parseInt(mapWidth);
    const scale = parseFloat(mapScale);

    Maps.update(mapID, {
      $set: {
        name: mapName,
        scale: scale,

        image: {
          width: width,
          height: height,
        },
      },
    });
  },

  /**
     * Removes the first map with the given name
     * @param mapName Common name of the map
     */
  'maps.removeMap'(mapName) {
    check(mapName, String);

    // Find the ID by its name
    const map = Maps.findOne({ name: mapName });

    // Remove that map by id
    if (map) {
      Maps.remove(map._id);
    }
  },

  /**
     * Adds a new node to this maps list of nodes in the database
     * @param map_id ID of the map to add this node to
     * @param latitude X position of the node
     * @param longitude Y position of the node
     */
  'maps.addNode'(mapID, latitude, longitude) {
    check(mapID, String);
    check(latitude, Number);
    check(longitude, Number);

    // Get the old list of nodes from the database
    const oldMap = Maps.findOne({ _id: mapID });
    const newNodesList = oldMap.nodes;

    // Update the list with a new node at the given latitude and longitude
    newNodesList.push({
      _id: Random.id(),
      lat: latitude,
      lng: longitude,
      name: 'Default Node',
      type: 'Hallway',

      gScore: -1,
      hScore: -1,
      fScore: -1,
      cameFrom: null,
    });

    // Update the Database by replacing the old set of nodes with the new one
    Maps.update(
      mapID,
      {
        $set:
            {
              nodes: newNodesList,
            },
      },
    );
  },

  /**
     * Deletes a node from the maps list of nodes
     * @param nodeID Mongo assigned ID of the node
     * @param mapID Mongo assigned ID of the map the node is in
     */
  'maps.removeNode'(nodeID, mapID) {
    check(nodeID, String);
    check(mapID, String);

    const map = Maps.findOne({ _id: mapID });

    let nodeList = map.nodes;
    let edgeList = map.edges;

    // Filters out any node and edge with a matching ID
    nodeList = nodeList.filter(function(node) {
      let keepNode = true;
      if (node._id === nodeID) {
        keepNode = false;
      }
      return keepNode;
    });
    edgeList = edgeList.filter(function(edge) {
      let keepEdge = true;
      if (edge.node1._id === nodeID || edge.node2._id === nodeID) {
        keepEdge = false;
      }
      return keepEdge;
    });

    // Update the database
    Maps.update(mapID, {
      $set: {
        nodes: nodeList,
        edges: edgeList,
      },
    });
  },

  /**
     * Updates a nodes position in the database
     * @param nodeID Mongo assigned ID of the node
     * @param mapID Mongo assigned ID of the map the node is in
     * @param newLat New X position
     * @param newLng New Y position
     */
  'maps.updateNodePos'(nodeID, mapID, newLat, newLng) {
    check(nodeID, String);
    check(mapID, String);
    check(newLat, Number);
    check(newLng, Number);

    // Get the old list of nodes and edges from the database
    const map = Maps.findOne({ _id: mapID });
    const nodeList = map.nodes;
    const edgeList = map.edges;

    // Find the node and edges associated with this node in this list and replace its values
    nodeList.forEach((node) => {
      if (node._id === nodeID) {
        node.lat = newLat;
        node.lng = newLng;
      }
    });
    edgeList.forEach((edge) => {
      if (edge.node1._id === nodeID) {
        edge.node1.lat = newLat;
        edge.node1.lng = newLng;
      } else if (edge.node2._id === nodeID) {
        edge.node2.lat = newLat;
        edge.node2.lng = newLng;
      }
    });

    // Update the Database by replacing the old set of nodes and edges with the new ones
    Maps.update(
      mapID,
      {
        $set: {
          nodes: nodeList,
          edges: edgeList,
        },
      },
    );
  },

  /**
     * Updates the name of the given node
     * @param nodeID Mongo assigned ID of the node
     * @param mapID Mongo assigned ID of the map
     * @param newName New name to change node to
     */
  'maps.updateNodeName'(nodeID, mapID, newName) {
    check(nodeID, String);
    check(mapID, String);
    check(newName, String);

    // Search out this node and update its name
    const nodes = Maps.findOne({ _id: mapID }).nodes;
    nodes.forEach((node) => {
      if (node._id === nodeID) { node.name = newName; }
    });

    // Update the DB
    Maps.update(mapID, {
      $set: {
        nodes,
      },
    });
  },

  /**
     * Returns a dictionary of node names and their map ID
     * @param nodeName
     */
  'maps.getNodeByName'(nodeName) {
    check(nodeName, String);

    const newNodeName = toTitleCase(nodeName);

    let foundNode = null;
    let foundMap = '';

    // Go through each map and its set of nodes to find a matching name,
    // incase multiple nodes have the same name
    const allMaps = Maps.find({}).fetch();
    allMaps.forEach((map) => {
      const nodes = map.nodes;
      nodes.forEach((node) => {
        if (toTitleCase(node.name) === newNodeName) {
          foundNode = node;
          foundMap = map._id;
        }
      });
    });

    return { node: foundNode, map: foundMap };
  },

  /**
     * Updates the type of the given node
     * @param nodeID Mongo assigned ID of the node
     * @param mapID Mongo assigned ID of the map
     * @param newName New name to change node to
     */
  'maps.updateNodeType'(nodeID, mapID, newType) {
    check(nodeID, String);
    check(mapID, String);
    check(newType, String);

    // Search out this node and update its type
    const nodes = Maps.findOne({ _id: mapID }).nodes;
    nodes.forEach((node) => {
      if (node._id === nodeID) { node.type = newType; }
    });

    // Update the DB
    Maps.update(
      mapID,
      {
        $set: {
          nodes,
        },
      },
    );
  },

  /**
     * Gets a list of all nodes in the database across all maps
     * @returns {Array} A list of nodes in the database
     */
  'maps.getAllNodes'() {
    let nodes = [];

    Maps.find({}).fetch().forEach((map) => {
      nodes = nodes.concat(map.nodes);
    });

    return nodes;
  },

  /**
     * A list of edges whose 'selected' field indicates whether or not this edge is on the main path
     * @param mapID Mongo ID of the map
     * @param edges A list of updated edges
     */
  'maps.highlightEdges'(mapID, edges) {
    check(mapID, String);
    check(edges, [Object]);

    Maps.update(mapID, {
      $set: {
        edges,
      },
    });
  },

  /**
     * Adds a new edge to this map
     * @param mapID Mongo assigned ID of the map
     * @param selectedNodes An array containing the list of nodes to create an edge between
     */
  'maps.addEdge'(mapID, selectedNodes, positions) {
    check(mapID, String);
    check(selectedNodes, Array);

    // Get the current list of edges or create one if it doesn't exist
    let edges = Maps.findOne({ _id: mapID }).edges;
    edges = edges || [];

    // Add the selected nodes and containing map as a new edge
    edges.push({
      node1: selectedNodes[0],
      node2: selectedNodes[1],
      positions,
      map: [mapID],
      selected: false,
      _id: Random.id(),
    });

    // Update the database with the new edges
    Maps.update(mapID, {
      $set: {
        edges,
      },
    });
  },

  /**
     * Creates an edge between two nodes on different (or the same, rendering ignored) map.
     * This only adds the edge
     * on the currently loaded map, but this is OK because the routing algorithm uses
     * a global list of edges.
     * @param mapID ID of the map to add the edge to
     * @param map2ID ID of the second map
     * @param node1 First node of the edge
     * @param node2 Second node of the edge
     */
  'maps.addCrossEdge'(mapID, map2ID, node1, node2) {
    check(mapID, String);
    check(map2ID, String);
    check(node1, Object);
    check(node2, Object);

    let edges = Maps.findOne({ _id: mapID }).edges;
    edges = edges || [];

    edges.push({
      node1,
      node2,
      positions: [],
      map: [mapID, map2ID],
      selected: false,
      _id: Random.id(),
    });

    Maps.update(mapID, {
      $set: {
        edges,
      },
    });
  },

  /**
     * Gets a list of all edges across all maps in the database
     * @returns {Array} A list of all edges in the database
     */
  'maps.getAllEdges'() {
    let edges = [];

    Maps.find({}).fetch().forEach((map) => {
      edges = edges.concat(map.edges);
    });

    return edges;
  },

  /**
     * Gets a list of locations that can be traveled to from all maps
     * @returns {Array} A list of names referring to node locations
     */
  'maps.getLocations'() {
    const locations = [];

    // Gather a list of locations
    const maps = Maps.find({}).fetch();
    maps.forEach((map) => {
      map.nodes.forEach((node) => {
        if (node.type === 'Location') {
          locations.push(toTitleCase(node.name));
        }
      });
    });

    // Return the list
    return locations;
  },

  /**
     * Gets a list of map names in the database
     * @return {Array} A list of map names
     */
  'maps.getMapNames'() {
    const names = [];

    // Gather a list of names
    const maps = Maps.find({}).fetch();
    maps.forEach((map) => {
      names.push(map.name);
    });

    // Return the list
    return names;
  },

  /**
     * Gets the first map found with the given name
     * @param mapName Name of the map
     * @returns {any} An object containing map data, or empty if none found
     */
  'maps.getMap'(mapName) {
    check(mapName, String);

    if (Array.isArray(mapName)) {
      return Maps.findOne({ name: mapName[0] });
    }
    return Maps.findOne({ name: mapName });
  },

  'maps.getMapByID'(mapID) {
    check(mapID, String);

    return Maps.findOne({ _id: mapID });
  },

  /**
     * Gets a list of image IDs corresponding to the list of mapIDs
     * @param mapIDs A list of map IDs
     * @returns {Array}
     */
  'maps.getImageIDs'(mapIDs) {
    check(mapIDs, [String]);

    const imageIDs = [];
    mapIDs.forEach((id) => {
      imageIDs.push(Maps.findOne({ _id: id }).image._id);
    });

    return imageIDs;
  },
});
