import React, { Component, PropTypes } from 'react';
import { Meteor } from 'meteor/meteor';
import { createContainer } from 'meteor/react-meteor-data';

import { Maps } from '../../api/collections/maps';
import { MapImages } from '../../api/collections/mapimages';
import { ImageOverlay, LayerGroup, LayersControl, Map, Rectangle } from 'react-leaflet';
import Toolbar from './Toolbar';
import Node from './Node';
import Edge from './Edge';
import { LatLng } from 'leaflet';

import AStar from '../../api/algorithms/astar';
import RoutingComponent from './RoutingComponent';
import L from 'leaflet';
import RouteGalleryContainer from './RouteGalleryContainer';

import Paper from 'material-ui/Paper';

const { BaseLayer, Overlay } = LayersControl;
const bounds = new L.LatLngBounds(new LatLng(0, 0), new LatLng(900, 600)); // to be replaced once component mounts
let boundsSet = false;
let zoomSet = false;

// App component - represents the whole app
class MapContainer extends Component {
  constructor(props) {
    super(props);

    // State of the application
    this.state = {
      bounds,
      minZoom: -1,
      maxZoom: 2,

      selectedTool: 'GENERAL',
      selectedNodes: [],

      currentPath: [],
      locations: [],
      globalEdges: [],
      globalNodes: [],

      routeDetails: null,
    };
  }

  /**
     * Currently updates the list of searchable locations if the database updates the 'Map' prop
     * @param nextProps New 'Map' object
     */
  componentWillReceiveProps(nextProps) {
    // We need to update the list of available locations
    Meteor.call('maps.getLocations', function(error, response) {
      if (!error) {
        this.setState({ locations: response });
      }
    }.bind(this));

    // Gather all of the edges in the system
    Meteor.call('maps.getAllEdges', function(err, res) {
      if (!err) {
        this.setState({ globalEdges: res });
      }
    }.bind(this));

    // Gather all of the nodes in the system
    Meteor.call('maps.getAllNodes', function(err, res) {
      if (!err) {
        // Set the cameFrom attribute of each node to null
        res.forEach((node) => {
          node.cameFrom = null;
        });

        this.setState({ globalNodes: res });
      } else {
        console.log('something went wrong getting all nodes');
        console.log(err);
      }
    }.bind(this));
  }

  /**
     * Resizes the map bounds to the image size whenever something changes to ensure we only click within map boundaries
     * Calculates the zoom boundaries once the map has been fit to the image bounds
     *
     * Sets Wrong Bounds --> Sets Zoom based on bounds --> Sets bounds based on new zoom
     */
  componentDidUpdate(nextProps, nextState) {
    // If the leaflet component has rendered, update the map bounds
    if (!!this.refs.map && !!this.props.Map && !!this.props.Map.image) {
      // Set the bounds if they're not set already once props have been created
      if (!boundsSet) {
        const leafletElt = this.refs.map.leafletElement;

        // Get the dimensions of the image from the map object
        const imageHeight = this.props.Map.image.height;
        const imageWidth = this.props.Map.image.width;

        // Calculate the coordinates relative to the image location offset by the map view
        const northEast = leafletElt.unproject([imageWidth, 0], this.state.maxZoom - 1);
        const southWest = leafletElt.unproject([0, imageHeight], this.state.maxZoom - 1);

        const newBounds = new L.LatLngBounds(southWest, northEast);

        // Set the new bounds
        this.setState({ bounds: newBounds });
        boundsSet = true;

        // Fit the zoom to the new bounds
        this.refs.map.leafletElement.fitBounds(newBounds);
      } else {
        // If the bounds have been set, see if we need to now calculate the zoom bounds
        if (!zoomSet) {
          const boundsZoom = this.refs.map.leafletElement.getBoundsZoom(this.state.bounds);

          // Set the state of min/max zooms
          this.setState({ maxZoom: boundsZoom + 2 });
          this.setState({ minZoom: boundsZoom - 1 });
          zoomSet = true;
        }
      }
    }

    // If the zoom has been calculated, dynamically set the min/max zoom of the map
    if (zoomSet) {
      this.refs.map.leafletElement.options.maxZoom = this.state.maxZoom;
      this.refs.map.leafletElement.options.minZoom = this.state.minZoom;
    }
  }

  /**
     * Recalculate the bounds next time this component is mounted
     */
  componentWillUnmount() {
    boundsSet = false;
    zoomSet = false;
  }

    /**
     * Switches tool to the given one
     * @param tool Name of tool to select
     */
    switchTool = (tool) => {
      this.setState({ selectedTool: tool });
    }

    /**
     * Clears the list of selected nodes. Triggers when edges are created, or nodes are removed.
     */
    clearSelectedNodes = () => {
      this.setState({ selectedNodes: [] });
    }

    /**
     * Sets all of the edges to not be highlighted so we can show a different route
     */
    clearHighlightedEdges = () => {
      this.props.Map.edges.forEach((edge) => {
        edge.selected = false;
      });
      Meteor.call('maps.highlightEdges', this.props.Map._id, this.props.Map.edges);
    }

    /**
     * Handles click events on the currently loaded leaflet map
     * @param event Click event
     */
    handleMapClick = (event) => {
      switch (this.state.selectedTool) {
        case 'GENERAL':
          console.log('general selected...');
          break;

        case 'ADD':
          // If the click was within the map bounds
          if (event.latlng.lat < this.state.bounds._northEast.lat && event.latlng.lat >= this.state.bounds._southWest.lat &&
                    event.latlng.lng < this.state.bounds._northEast.lng && event.latlng.lng >= this.state.bounds._southWest.lng) {
            // Create a new node!
            Meteor.call('maps.addNode', this.props.Map._id, event.latlng.lat, event.latlng.lng);
          } else {
            console.log('Point not within map bounds!');
          }
          break;

        case 'EDGE':
          console.log('edge selected...');
          break;

        default:
          console.log('No tool currently selected');
      }
    }

    /**
     * Runs the AStar algorithm on the given nodes and highlights the route
     */
    handleRoute = (startName, endName) => {
      // Clear the previous path
      this.clearHighlightedEdges();

      let startNode = null;
      let endNode = null;
      Meteor.call('maps.getNodeByName', startName, function(err, resStart) {
        if (!err) {
          startNode = resStart.node;

          Meteor.call('maps.getNodeByName', endName, function(err, resEnd) {
            if (!err) {
              endNode = resEnd.node;

              if (startNode && endNode) {
                const astar = new AStar(
                  this.state.globalNodes,
                  this.state.globalEdges,
                  startNode,
                  endNode,
                );

                console.log(`Running AStart from node ${startNode.name}(${startNode._id}) to ${endNode.name}(${endNode._id})`);

                // Run the algorithm to find the best path and highlight it if it exists
                if (astar.runAlgorithm()) {
                  console.log('Found a path!');
                  const reconstruction = astar.reconstructPath(); // [0] - path, [1] - edges

                  // Print out the path in order
                  console.log(reconstruction[0].map(function (node) {
                    return node._id;
                  }).join(' --> '));

                  // Highlight path edges
                  /* this.state.globalEdges.forEach((edge) => {
                                    reconstruction[1].forEach((edgeRec) => {
                                        if (edgeRec._id === edge._id) {
                                            edge.selected = true;
                                        }
                                    });
                                }); */

                  // Meteor.call('maps.highlightEdges', this.props.Map._id, this.state.globalEdges);

                  // Do the stuff the gallery should do
                  const mapIDs = [];
                  reconstruction[1].forEach((edge) => {
                    // if (edge.map.length > 1) {
                    edge.map.forEach((id) => {
                      if (!mapIDs.includes(id)) { mapIDs.push(id); }
                    });
                    // }
                  });

                  // For every map ID, get the map object and update its selected edges
                  mapIDs.forEach((map) => {
                    Meteor.call('maps.getMapByID', map, function(err, res) {
                      console.log(res);
                      res.edges.forEach((edge) => {
                        reconstruction[1].forEach((edgeRec) => {
                          if (edge._id === edgeRec._id) {
                            edge.selected = true;
                          }
                        });
                      });

                      Meteor.call('maps.highlightEdges', map, res.edges);
                    });
                  });

                  console.log(mapIDs);

                  if (mapIDs.length > 1) {
                    // Use the new mapIDs to gather new image links, otherwise set only the mapIDs list
                    Meteor.call('maps.getImageIDs', mapIDs, function (err, res) {
                      if (!err) {
                        Meteor.call('mapimages.getImageLinks', res, function (linkErr, linkRes) {
                          if (!err) {
                            console.log(linkRes);
                            this.setState({ routeDetails: [mapIDs, linkRes] });
                          } else { console.log(linkErr); }
                        }.bind(this));
                      } else {
                        console.log(err);
                      }
                    }.bind(this));
                  }

                  // END THE STUFF INSIDE GALLERY
                } else {
                  console.log('AStar was unable to find a path');
                }
              }
            } else {
              console.log(err);
            }
          }.bind(this));
        } else {
          console.log(err);
        }
      }.bind(this));
    }

    createCrossMapEdge = (node1, node2_name) => {
      /**
         * 1. Get the node ID of node2
         * 2. Get the map of node2
         * 3. Get the current map
         * 4. Create the edge
         */
      Meteor.call('maps.getNodeByName', node2_name, function(err, res) {
        if (!err) {
          console.log(res);
          const node2 = res.node;
          const map2ID = res.map;

          Meteor.call('maps.addCrossEdge', this.props.Map._id, map2ID, node1, node2);
        }
      }.bind(this));
    }

    /**
     * Called by the Node class. Selects this node object for creating edges.
     * @param node Node object selected
     */
    selectNode = (node) => {
      switch (this.state.selectedTool) {
        case 'EDGE':
          var currentNodes = this.state.selectedNodes;
          var nodeExisted = false;

          // Check if the node has been selected before, and if so deselect it
          currentNodes.forEach((existingNode) => {
            if (existingNode._id === node._id) {
              nodeExisted = true;
              currentNodes.splice(currentNodes.indexOf(node), 1);
            }
          });

          // If the node did not exist in the list before, begin adding process
          if (!nodeExisted) {
            // Add the node to the set of selected nodes
            currentNodes.push(node);
            this.setState({ selectedNodes: currentNodes });

            // Check if there have been 2 selections
            if (this.state.selectedNodes.length === 2) {
              // Get position
              const positions = [new LatLng(this.state.selectedNodes[0].lat, this.state.selectedNodes[0].lng),
                new LatLng(this.state.selectedNodes[1].lat, this.state.selectedNodes[1].lng)];

              // We need to create an edge
              Meteor.call('maps.addEdge', this.props.Map._id, this.state.selectedNodes, positions);

              // We want to deselect all nodes
              this.clearSelectedNodes();
            }
          } else {
            console.log(`Node ${node._id} has already been selected and is now deselected`);
          }
          break;

        default:
          console.log('EDGE tool not selected, can not select node for connection');
      }
    }

    /**
     * Renders the nodes associated with the current map
     * @returns {Array} Array of React components representing nodes
     */
    renderNodes() {
      const nodes = this.props.Map.nodes;

      return nodes ? nodes.map(node => (
        <Node
          key={node._id}
          node={node}
          mapID={this.props.Map._id}
          locations={this.state.locations}

          onSelect={this.selectNode}
          onDelete={this.clearSelectedNodes}
          onCreateConnection={this.createCrossMapEdge}
        />
      )) : console.log('No map loaded, can not render any nodes.');
    }

    /**
     * Renders the edges associated with the current map
     * @returns {Array} Array of react components representing edges
     */
    renderEdges() {
      const edges = this.props.Map.edges;

      return edges ? edges.map((edge) => {
        if (edge.map.length > 1) { return; } // skip cross map edges
        return (
          <Edge
            key={edge._id}
            edge={edge}
          />
        );
      }) : console.log('No map loaded, can not render any edges.');
    }

    /**
     * Renders the map editor component
     * @returns {XML}
     */
    render() {
      return (
        <Paper style={{ width: '100%', height: '100%' }}>
          <Map
            className="loaded-map"
            bounds={this.state.bounds}
            onClick={this.handleMapClick}
            ref="map"
          >
            <Rectangle bounds={this.state.bounds} />
            {/* <RoutingComponent
                    locations={this.state.locations}
                    onRoute={this.handleRoute}
                mapID={!!this.props.Map._id ? this.props.Map._id : ""}/> */}
            <LayersControl position="topright">
              <Overlay checked name="Image Overlay">
                <LayerGroup>
                  <ImageOverlay
                    checked
                    name="Image Overlay"
                    bounds={this.state.bounds}
                    attribution="Worcester Polytechnic Institute MQP"
                    url={this.props.mapImage}
                  />
                </LayerGroup>
              </Overlay>
              <Overlay checked name="Nodes Overlay">
                <LayerGroup>
                  /*{this.renderNodes()}*/
                </LayerGroup>
              </Overlay>

              <Overlay checked name="Edges Overlay">
                <LayerGroup>
                  /*{this.renderEdges()}*/
                </LayerGroup>
              </Overlay>

              <Toolbar onSwitch={this.switchTool} />
              {this.state.routeDetails ? <RouteGalleryContainer
                routeDetails={this.state.routeDetails}
                loadMap={this.props.loadMap}
              /> : ''}
            </LayersControl>
          </Map>
        </Paper>
      );
    }
}

/**
 * A set of proptypes for the MapContainer
 * @type {{Map: *}} Map loaded from the database
 */
MapContainer.propTypes = {
  Map: PropTypes.object.isRequired,
  loadMap: PropTypes.func.isRequired,
};

/**
 * Meteor default container for this component. Will subscribe to the database and populate
 * the "Map" prop when the container is created.
 * @param mapName Could optionally receive a mapName if triggered by loadMapPopup
 */
export default createContainer(({ mapName }) => {
  Meteor.subscribe('maps');
  Meteor.subscribe('mapimages');

  const mapToLoad = mapName ? Maps.findOne({ name: mapName }, { sort: { updated: -1, limit: 1 } })
    : Maps.findOne({}, { sort: { updated: -1, limit: 1 } });

  return {
    Map: (mapToLoad || {}), // initially empty, but will be updated after the first render
    mapImage: (mapToLoad ? MapImages.findOne({ _id: mapToLoad.image._id }).link() : ''),
  };
}, MapContainer);
