import React, {Component, PropTypes} from 'react';
import {Marker, Popup} from "react-leaflet";
import {Meteor} from 'meteor/meteor';

import PopupReact from 'react-popup';
import {Typeahead} from "react-bootstrap-typeahead";
import {toTitleCase} from "../../api/algorithms/strings";

export default class Node extends Component {
    constructor(props){
        super(props);

        // Initial state of a Node
        this.state = {
                center: {
                    lat: this.props.node.lat,
                    lng: this.props.node.lng
                },
                marker: {
                    lat: this.props.node.lat,
                    lng: this.props.node.lng
                },

                zoom: 1,
                draggable: true,
                name: this.props.node.name,

                crossNodeName: '',
                hideDropDown: false,
            }
    }

    /**
     * Updates the position of the node and sets its state to account for this. Then updates the database with the
     * new positions which should update the map editor.
     */
    updatePosition = () => {
        // Get the position from the marker being rendered
        const {lat, lng} = this.refs.marker.leafletElement.getLatLng();

        // Set the position of the marker
        this.setState({
            marker: {lat, lng},
        });

        // Update the database for this node
        Meteor.call('maps.updateNodePos', this.props.node._id, this.props.mapID, lat, lng);
    }

    /**
     * Deletes the node from the database therefore removing the node from the list of nodes on this map and
     * no longer rendering
     */
    handleDelete = () => {
        // Update the database
        Meteor.call('maps.removeNode', this.props.node._id, this.props.mapID);

        // Remove all selected nodes
        this.props.onDelete();
    }

    /**
     * Selects this node for edge creation and passes it back to the mapcontainer through onSelect(node)
     */
    handleClick = () => {
        this.props.onSelect(this.props.node);
    }

    /**
     * Handles a change in node name
     * @param event Event containing the new text
     */
    handleNameChange = (e) => {
        const text = e.target.value.trim();
        this.setState({name: text});

        Meteor.call('maps.updateNodeName', this.props.node._id, this.props.mapID, text);
    }

    /**
     * Handles a change in node type
     * @param type Event containing the new type information
     */
    handleChangeType = (type) => {
        const newType = type.target.value.trim();
        Meteor.call('maps.updateNodeType', this.props.node._id, this.props.mapID, newType);
    }

    /**
     * Shows the typeahead dropdown
     */
    showDropDown = () => {
        this.setState({hideDropDown: false});
    }

    /**
     * Handles Option Selection if one is clicked instead of the name being typed
     */
    selectOption = (value) => {
        this.setState({crossNodeName: value[0], hideDropDown: true});
    }

    createEdge = () => {
        this.props.onCreateConnection(this.props.node, this.state.crossNodeName);
    }

    /**
     * Renders the node component to the screen
     * @returns {XML}
     */
    render() {
        var nodeName = toTitleCase(this.props.node.name);

        return(
            <Marker
                position={[this.state.marker.lat, this.state.marker.lng]}
                draggable={this.state.draggable}
                ondragend={this.updatePosition}
                onClick={Meteor.user() ? this.handleClick : ''}
                ref="marker">

                {Meteor.user() ?
                <Popup minWidth={90}>
                    <span className="node-span">
                        <div>
                            Node Name:
                            <input onChange={this.handleNameChange} type="text" value={this.props.node.name}/>
                        </div>
                        <div>
                            Node Type:
                            <select value={this.props.node.type ? this.props.node.type : "Location"} onChange={this.handleChangeType}>
                                <option value="Location">Location</option>
                                <option value="Hallway">Hallway</option>
                            </select>
                        </div>
                        <div>
                            Connection:
                            <Typeahead
                                onChange={this.selectOption}
                                onFocus={this.showDropDown}
                                type="text"
                                placeholder={'Enter name of node to connect...'}
                                options={this.state.hideDropDown ? [] : this.props.locations}/>
                            <button onClick={this.createEdge}>Connect</button>
                        </div>
                        <br/>
                        <div className="center">
                            <button style={{backgroundColor: "tomato"}} onClick={this.handleDelete}>Delete Node</button>
                        </div>
                    </span>
                </Popup>
                    :
                    <Popup minWidth={nodeName.length + 5}>
                        <span className="node-user-span">
                            <div className="node-user-span">
                                {nodeName}
                            </div>
                        </span>
                    </Popup>}
            </Marker>
        )
    }
}

Node.propTypes={
    node: PropTypes.object.isRequired,
    mapID: PropTypes.string.isRequired,

    onSelect: PropTypes.func,
    onDelete: PropTypes.func,
    onCreateConnection: PropTypes.func,
}