import React, { Component } from 'react';
import PropTypes from 'prop-types';
import Paper from 'material-ui/Paper';
import { List, ListItem } from 'material-ui/List';
import { Meteor } from 'meteor/meteor';

import AddMapDialog from './dialogs/AddMapDialog.jsx';
import LoadMapDialog from './dialogs/LoadMapDialog.jsx';
import EditMapDialog from './dialogs/EditMapDialog.jsx';
import RemoveMapDialog from './dialogs/RemoveMapDialog.jsx';
import SidebarHeader from './SidebarHeader.jsx';
import { toTitleCase } from '../../api/algorithms/strings';
import PluginViewComponent from '../pluginmanager/ui/PluginViewComponent';
import MapContainer from '../mapeditor/MapContainer.jsx';
import SchoolConfig from '../pluginmanager/plugins/schooldirectory/SchoolConfig.jsx';

const sidebarHeaderStyles = { paddingRight: '0px', width: '100%' };
const sidebarListStyles = { paddingRight: '0px', width: '100%', height: '100%' };

class SideBar extends Component {
  constructor(props) {
    super(props);

    this.state = {
      addOpen: false,
      loadOpen: false,
      editOpen: false,
      removeOpen: false,
    };
  }

  toggleAdd = () => { this.setState({ addOpen: !this.state.addOpen }); }
  toggleLoad = () => { this.setState({ loadOpen: !this.state.loadOpen }); }
  toggleEdit = () => { this.setState({ editOpen: !this.state.editOpen }); }
  toggleRemove = () => { this.setState({ removeOpen: !this.state.removeOpen }); }
  loadSchool = () => { this.props.loadPlugin(<SchoolConfig />); };
  loadHospital = () => { console.log('This would load the hospital'); };
  loadPluginManager = () => {
    if (Meteor.user()) {
      this.props.loadPlugin(<PluginViewComponent />);
    }
  };
  loadMapEditor = () => { this.props.loadPlugin(<MapContainer loadMap={this.props.loadMap} />); };

    renderPlugins = () => {
      const pluginsList = [];

      this.props.plugins.forEach(function(plugin) {
        const name = toTitleCase(plugin.name);

        if (plugin.installed) {
          if (plugin.name === 'SCHOOL_DIRECTORY') {
            pluginsList.push(<ListItem key={plugin.name} primaryText={name} onClick={this.loadSchool} />);
          } else if (plugin.name === 'HOSPITAL_DIRECTORY') {
            pluginsList.push(<ListItem key={plugin.name} primaryText={name} onClick={this.loadHospital} />);
          } else {
            pluginsList.push(<ListItem key={plugin.name} primaryText={name} />);
          }
        }
      }.bind(this));

      return pluginsList;
    }

    renderMapOptions = () => [
      <ListItem key="add-map" primaryText="Add New Map" onClick={this.toggleAdd} />,
      <ListItem key="load=map" primaryText="Load Map" onClick={this.toggleLoad} />,
      <ListItem key="edit-map" primaryText="Edit Map" onClick={this.toggleEdit} />,
      <ListItem key="remove-map" primaryText="Remove Map" onClick={this.toggleRemove} />,
    ]

    render() {
      const pluginsToRender = Meteor.user() ? this.renderPlugins() : [];
      const mapOptions = Meteor.user() ? this.renderMapOptions() : [];

      let dialog;
      if (this.state.addOpen) {
        dialog = <AddMapDialog onClose={this.toggleAdd} isOpen={this.state.addOpen} />;
      } else if (this.state.loadOpen) {
        dialog = <LoadMapDialog onClose={this.toggleLoad} isOpen={this.state.loadOpen} />;
      } else if (this.state.editOpen) {
        dialog = <EditMapDialog onClose={this.toggleEdit} isOpen={this.state.editOpen} />;
      } else if (this.state.removeOpen) {
        dialog = <RemoveMapDialog onClose={this.toggleRemove} isOpen={this.state.removeOpen} />;
      } else {
        dialog = <div />;
      }

      return (
        <div className="sidebar-components">
          <SidebarHeader style={sidebarHeaderStyles} />
          <Paper>
            <List style={sidebarListStyles} >
              <ListItem primaryText="Maps" onClick={this.loadMapEditor} autoGenerateNestedIndicator nestedItems={mapOptions} />
              <ListItem primaryText="Plugins" onClick={this.loadPluginManager} autoGenerateNestedIndicator nestedItems={pluginsToRender} />
            </List>
          </Paper>
          {dialog}

        </div>
      );
    }
}

SideBar.propTypes = {
  plugins: PropTypes.array.isRequired,
  loadPlugin: PropTypes.func.isRequired,
  loadPluginConfig: PropTypes.func.isRequired,
  loadMap: PropTypes.func.isRequired,
};

export default SideBar;
