import React, {Component, PropTypes} from 'react';
import {Meteor} from 'meteor/meteor';
import { createContainer } from 'meteor/react-meteor-data';

import {Maps} from "../../api/collections/maps";
import {MapImages} from "../../api/collections/mapimages";
import RoutingComponent from "../mapeditor/RoutingComponent";
import RouteGalleryContainer from "../mapeditor/RouteGalleryContainer";
import L from "leaflet";
import AStar from '../../../imports/api/algorithms/astar';
import {LatLng} from "leaflet";
import {ImageOverlay, Map} from "react-leaflet";
import Node from '../../ui/mapeditor/Node';
import Edge from "../mapeditor/Edge";

const bounds = new L.LatLngBounds(new LatLng(0, 0), new LatLng(750, 620)); // to be replaced once component mounts
var boundsSet = false;
var zoomSet = false;

class UserMapComponent extends Component {

    constructor(props) {
        super(props);

        // State of the application
        this.state = {
            bounds: bounds,
            minZoom: null,
            maxZoom: 20,

            currentPath: [],
            locations: [],
            globalEdges: [],
            globalNodes: [],

            routeDetails: null,
        }
    }

    /**
     * Resizes the map bounds to the image size whenever something changes to ensure we only click within map boundaries
     * Calculates the zoom boundaries once the map has been fit to the image bounds
     *
     * Sets Wrong Bounds --> Sets Zoom based on bounds --> Sets bounds based on new zoom
     */
    componentDidUpdate(nextProps, nextState) {
        // If the leaflet component has rendered, update the map bounds
        if (!!this.refs.map && !!this.props.Map && !!this.props.Map.image) {
            // Set the bounds if they're not set already once props have been created
            if (!boundsSet) {
                var leafletElt = this.refs.map.leafletElement;

                // Get the dimensions of the image from the map object
                var imageHeight = this.props.Map.image.height;
                var imageWidth = this.props.Map.image.width;

                // Calculate the coordinates relative to the image location offset by the map view
                var northEast = leafletElt.unproject([0, imageHeight], this.state.maxZoom - 1);
                var southWest = leafletElt.unproject([imageWidth, 0], this.state.maxZoom - 1);

                // Set the new bounds
                this.setState({bounds: new L.LatLngBounds(southWest, northEast)});
                boundsSet = true;

                // Fit the zoom to the new bounds
                this.refs.map.leafletElement.fitBounds(this.state.bounds);
            } else {

                // If the bounds have been set, see if we need to now calculate the zoom bounds
                if (!zoomSet) {
                    var boundsZoom = this.refs.map.leafletElement.getBoundsZoom(this.state.bounds);

                    // Set the state of min/max zooms
                    this.setState({maxZoom: boundsZoom + 2});
                    this.setState({minZoom: boundsZoom - 1});
                    zoomSet = true;
                }
            }
        }

        // If the zoom has been calculated, dynamically set the min/max zoom of the map
        if (zoomSet) {
            this.refs.map.leafletElement.options.maxZoom = this.state.maxZoom;
            this.refs.map.leafletElement.options.minZoom = this.state.minZoom;
        }
    }

    /**
     * Currently updates the list of searchable locations if the database updates the 'Map' prop
     * @param nextProps New 'Map' object
     */
    componentWillReceiveProps(nextProps){
        // We need to update the list of available locations
        Meteor.call('maps.getLocations', function(error, response){
            if(!error){
                this.setState({locations: response});
            }
        }.bind(this));

        // Gather all of the edges in the system
        Meteor.call('maps.getAllEdges', function(err, res){
            if(!err){
                this.setState({globalEdges: res})
            }
        }.bind(this));

        // Gather all of the nodes in the system
        Meteor.call('maps.getAllNodes', function(err, res){
            if(!err){
                // Set the cameFrom attribute of each node to null
                res.forEach((node) => {
                    node.cameFrom = null;
                });

                this.setState({globalNodes: res});
            } else {
                console.log('something went wrong getting all nodes');
                console.log(err);
            }
        }.bind(this));
    }

    /**
     * Renders the nodes associated with the current map
     * @returns {Array} Array of React components representing nodes
     */
    renderNodes() {
        let nodes = this.props.Map.nodes;

        return !!nodes ? nodes.map((node) => {
            if(node.type !== 'Location'){return}
            return(
                <Node
                    key={node._id}
                    node={node}
                    mapID={this.props.Map._id}
                    locations={this.state.locations}
                />
            )
        }) : console.log('No map loaded, can not render any nodes.');

    }

    renderHighlightedEdges(){
        let edges = this.props.Map.edges;

        return !!edges ? edges.map((edge) => {
            if(edge.map.length > 1 || !edge.selected){ return } // skip cross map edges or unselected edges
            return (
                <Edge
                    key={edge._id}
                    edge={edge}
                />
            )
        }) : console.log('No map loaded, can not render any edges.');
    }

    /**
     * Runs the AStar algorithm on the given nodes and highlights the route
     */
    handleRoute = (startName, endName) => {
        // Clear the previous path
        this.clearHighlightedEdges();

        var startNode = null;
        var endNode = null;
        Meteor.call('maps.getNodeByName', startName, function(err, resStart){
            if(!err){
                startNode = resStart.node;

                Meteor.call('maps.getNodeByName', endName, function(err, resEnd){
                    if(!err){
                        endNode = resEnd.node;

                        if(startNode && endNode) {
                            const astar = new AStar(
                                this.state.globalNodes,
                                this.state.globalEdges,
                                startNode,
                                endNode);

                            console.log('Running AStart from node ' + startNode.name + '(' + startNode._id + ') to ' + endNode.name + '(' + endNode._id + ')');

                            // Run the algorithm to find the best path and highlight it if it exists
                            if (astar.runAlgorithm()) {
                                console.log('Found a path!');
                                var reconstruction = astar.reconstructPath(); // [0] - path, [1] - edges

                                // Print out the path in order
                                console.log(reconstruction[0].map(function (node) {
                                    return node._id;
                                }).join(" --> "));

                                // Do the stuff the gallery should do
                                var mapIDs = [];
                                reconstruction[1].forEach((edge) => {
                                    edge.map.forEach((id) => {
                                        if(!mapIDs.includes(id))
                                            mapIDs.push(id);
                                    });
                                });

                                // For every map ID, get the map object and update its selected edges
                                mapIDs.forEach((map) => {
                                    Meteor.call('maps.getMapByID', map, function(err, res){
                                        console.log(res);
                                        res.edges.forEach((edge) => {
                                            reconstruction[1].forEach((edgeRec) => {
                                                if(edge._id === edgeRec._id){
                                                    edge.selected = true;
                                                }
                                            })
                                        });

                                        Meteor.call('maps.highlightEdges', map, res.edges);
                                    });
                                });

                                console.log(mapIDs);

                                if(mapIDs.length > 1) {
                                    // Use the new mapIDs to gather new image links, otherwise set only the mapIDs list
                                    Meteor.call('maps.getImageIDs', mapIDs, function (err, res) {
                                        if (!err) {
                                            Meteor.call('mapimages.getImageLinks', res, function (linkErr, linkRes) {
                                                if (!err) {
                                                    console.log(linkRes);
                                                    this.setState({routeDetails: [mapIDs, linkRes]})
                                                }
                                                else
                                                    console.log(linkErr);
                                            }.bind(this));
                                        } else {
                                            console.log(err);
                                        }
                                    }.bind(this));
                                }

                                // END THE STUFF INSIDE GALLERY
                            }
                            else {
                                console.log('AStar was unable to find a path');
                            }
                        }
                    } else {
                        console.log(err);
                    }
                }.bind(this))
            } else {
                console.log(err);
            }
        }.bind(this));
    }

    handleMapClick = () => {

    }

    /**
     * Sets all of the edges to not be highlighted so we can show a different route
     */
    clearHighlightedEdges = () => {
        this.props.Map.edges.forEach((edge) => {
            edge.selected = false;
        });
        Meteor.call('maps.highlightEdges', this.props.Map._id, this.props.Map.edges);
    }

    render() {
        return (
            <div>
                <RoutingComponent
                    locations={this.state.locations}
                    onRoute={this.handleRoute}
                    mapID={!!this.props.Map._id ? this.props.Map._id : ""}/>
                <Map className="loaded-map"
                     bounds={this.state.bounds}
                     onClick={this.handleMapClick}
                     ref="map"
                >
                    <ImageOverlay
                        checked
                        name="Image Overlay"
                        bounds={this.state.bounds}
                        attribution="Worcester Polytechnic Institute MQP"
                        url={this.props.mapImage}/>

                    {this.renderNodes()}
                    {this.renderHighlightedEdges()}

                        {this.state.routeDetails ? <RouteGalleryContainer
                            routeDetails={this.state.routeDetails}
                            loadMap={this.props.loadMap}/> : ''}
                </Map>
            </div>
        )
    }
}

UserMapComponent.propTypes = {
    Map: PropTypes.object.isRequired,
    loadMap: PropTypes.func.isRequired,
}

export default createContainer((props) => {
    Meteor.subscribe('maps');
    Meteor.subscribe('mapimages');

    const mapToLoad = props.mapName ? Maps.findOne({name: props.mapName}) : Maps.findOne({}, {sort: {updated: -1, limit: 1}});

    return {
        Map: mapToLoad ? mapToLoad : {},
        mapImage: (mapToLoad ? MapImages.findOne({_id: mapToLoad.image._id}).link() : ''),
    };
}, UserMapComponent);